// Copyright (C) 2016 Jochen Becher
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#pragma once

#include "qmt/controller/references.h"
#include "qmt/infrastructure/qmt_global.h"

namespace qmt {

class MElement;

class QMT_EXPORT MReferences : public References<MElement>
{
};

} // namespace qmt
