// Copyright (C) 2016 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#include "markup.h"

/*!
    \class BinEditor::Markup
    \brief The Markup class defines the range of the binary editor.

    Used for displaying class layouts by the debugger.

    \note Must not have linkage - used for soft dependencies.

    \sa Debugger::Internal::MemoryAgent
*/
