// Copyright (C) 2016 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#pragma once

namespace Analyzer {
namespace Icons {

const char SETTINGSCATEGORY_ANALYZER[] = ":/images/settingscategory_analyzer.png";

} // namespace Icons
} // namespace Analyzer
